# Copyright 3D Control Systems, Inc. All Rights Reserved 2017-2019.
# Built in San Francisco.

# This software is distributed under a commercial license for personal,
# educational, corporate or any other use.
# The software as a whole or any parts of it is prohibited for distribution or
# use without obtaining a license from 3D Control Systems, Inc.

# All software licenses are subject to the 3DPrinterOS terms of use
# (available at https://www.3dprinteros.com/terms-and-conditions/),
# and privacy policy (available at https://www.3dprinteros.com/privacy-policy/)

import logging
import os
import requests
import pathlib
import sys
import tempfile
import threading
import time
import gzip

import config
import log
import paths


class Downloader(threading.Thread):

    CONNECTION_TIMEOUT = 6
    MAX_RETRIES = 15
    CHUNK_SIZE = 1024 * 1024 # 1MB

    def __init__(self, parent, url, callback, is_zip):
        self.percent = 0
        self.url = url
        self.is_zip = is_zip
        self.parent = parent
        self.callback = callback
        self.cancel_flag = False
        self.logger = logging.getLogger(__name__)
        threading.Thread.__init__(self, name="Downloader", daemon=True)

    @log.log_exception
    def run(self):
        self.logger.info('Starting downloading')
        downloaded_filename = self.download()
        if downloaded_filename:
            self.execule_callback(downloaded_filename)
        if self.cancel_flag:
            self.logger.info('Cancel command was received after printing start in downloading thread')
            try:
                self.parent.printer.cancel()
            except AttributeError:
                pass
        self.logger.info('Downloading finished')

    def execule_callback(self, f):
        if self.cancel_flag:
            self.logger.info('Cancel command received')
            self.cancel_flag = False
        elif not f:
            self.parent.register_error(67, "Unknown download error", is_blocking=True)
        else:
            self.callback(f)

    def download(self):
        self.logger.info("Downloading from " + self.url)
        if self.is_zip:
            suffix = ".zip"
        else:
            suffix = ".gcode"
        self.tmp_file = tempfile.NamedTemporaryFile(mode='wb', dir=paths.DOWNLOAD_FOLDER,
                delete=False, prefix='3dprinteros-', suffix=suffix)
        tmpfile_name = self.tmp_file.name
        retry = 0
        resume_byte_pos = 0 
        while retry < self.MAX_RETRIES and not self.parent.stop_flag and not self.cancel_flag:
            if retry:
                self.logger.warning("Download retry/resume N%d" % retry)
            self.logger.info("Connecting to " + self.url)
            headers = { 'Accept-Encoding': 'identity, deflate, compress, gzip',
                     'Accept': '*/*', 'User-Agent': 'python-requests/{requests.__version__}'}
            if resume_byte_pos:
                headers['Range'] = 'bytes=%d-' % resume_byte_pos
            try:
                response = requests.get(self.url, headers = headers, stream=True, timeout = self.CONNECTION_TIMEOUT)
            except Exception as e:
                response = None
                self.parent.register_error(65, "Unable to open download link: " + str(e), is_blocking=False)
            else:
                self.logger.info("Successful connection to " + self.url)
                download_length = int(response.headers.get('content-length', 0))
                self.logger.debug(f'Download headers: {response.headers}')
                # download_filename = request.headers.get('content-disposition', "").split("filename=")[1]
                if download_length:
                    downloaded_size = self.download_chunks(response, download_length)
                    resume_byte_pos += downloaded_size
                    self.logger.info("Downloaded %d bytes" % resume_byte_pos)
                    if downloaded_size == download_length:
                        self.tmp_file.close()
                        return tmpfile_name
                    elif downloaded_size > download_length:
                        self.parent.register_error(66, 'Error while downloading: data is corrupted', is_blocking=True)
                        break
            finally:
                if response:
                    response.close()
                retry += 1
                time.sleep(1)
        self.tmp_file.close()
        try:
            os.remove(tmpfile_name)
        except:
            pass

    def download_chunks(self, response, download_length):
        # Taking +1 byte with each chunk to compensate file length till less than 100 bytes when dividing by 100
        percent_length = download_length // 100 + 1
        downloaded_size = 0
        try:
            for chunk in response.iter_content(percent_length):
                if self.cancel_flag or self.parent.stop_flag:
                    self.logger.info('Download canceled')
                    break
                self.tmp_file.write(chunk)
                self.percent += 1
                total_size = response.raw.tell()
                self.logger.info('File downloading : %d%%' % self.percent)
        except Exception as e:
            self.logger.exception(e)
            self.parent.register_error(66, 'Error while downloading: ' + str(e), is_blocking=False)
        finally:
            return downloaded_size

    def cancel(self):
        self.cancel_flag = True

    def get_percent(self):
        return self.percent
